
#include <stdio.h>
#include <string.h>

#define MAX_PALETTE     256 // entries
#define MAX_SCANLINE 131072 // bytes

#define ERR_OPEN  "Error opening file"
#define ERR_READ  "Error reading file"
#define ERR_WRITE "Error writing file"

#pragma pack(push,1)
typedef struct {
	__int16 bm;
	__int32 fsize,reserved,dataofs,headsize,xsize,ysize;
	__int16 planes,bpp;
	__int32 compress,mapsize,xres,yres,palsize,palimp;
} BMPhead;
#pragma pack(pop)

int roundup_bits(int n) {
	return ((n+31)/32)*4;
}

void bitcopy(unsigned char *dst,const unsigned char *src,int dst_offset,int src_offset,int len) {
  int dst_int = dst_offset / 8 , src_int = src_offset / 8;
  int dst_mod = dst_offset % 8 , src_mod = src_offset % 8;
  int len_mid = (len + dst_mod) / 8 - 1;
  int shift = (src_mod - dst_mod) % 8;
  if (shift<0) { shift += 8; src_int--; }
  int inv_shift = 8 - shift;

  unsigned char mask1 = (0xFF>>dst_mod);
  unsigned char mask2 = (0xFF<<8-(dst_offset+len)%8);
  unsigned char byte1 = mask1 & ((src[src_int]<<shift) | (src[src_int+1]>>inv_shift));
  unsigned char byte2 = mask2 & ((src[src_int+len_mid+1]<<shift) | (src[src_int+len_mid+2]>>inv_shift));

  if (len_mid>=0) {
    dst[dst_int] = (~mask1 & dst[dst_int]) | byte1;
    for (int i=1; i<=len_mid; i++) {
      dst[dst_int+i] = (src[src_int+i]<<shift) | (src[src_int+i+1]>>inv_shift);
    }
    dst[dst_int+len_mid+1] = (~mask2 & dst[dst_int+len_mid+1]) | byte2;
  }
  else {
    dst[dst_int] = ~(mask1 & mask2) & dst[dst_int] | (byte1 & byte2);
  }
}

bool crop(char *name_in, char *name_out, int x1, int y1, int x2, int y2) {
	BMPhead head_in,head_out;
	char palette[4*MAX_PALETTE];
	unsigned char buffer_in[MAX_SCANLINE];
	unsigned char buffer_out[MAX_SCANLINE];
	int read,written;
	FILE* fp_in = NULL;
	FILE* fp_out = NULL;
	memset(&palette,0,sizeof(palette));

	try {
		fp_in = fopen(name_in,"rb");
		if (!fp_in) throw ERR_OPEN;

		printf("Reading %s ...\n",name_in);

		read = fread(&head_in,1,sizeof(head_in),fp_in);
		if (read<sizeof(head_in)) throw ERR_READ;

		int palsize_bytes = head_in.dataofs-sizeof(head_in);

		read = fread(&palette,1,palsize_bytes,fp_in);
		if (read<palsize_bytes) throw ERR_READ;

		head_out=head_in;
		head_out.xsize = x2-x1;
		head_out.ysize = y2-y1;
		int scanline_in_bits  = head_in.xsize*head_in.bpp;
		int scanline_out_bits = head_out.xsize*head_out.bpp;
		int rounded_scanline_in = roundup_bits(scanline_in_bits);
		int rounded_scanline_out = roundup_bits(scanline_out_bits);
		head_out.fsize = head_out.dataofs+head_out.ysize*rounded_scanline_out;

		fp_out = fopen(name_out,"wb");
		printf("Writing %s ...\n",name_out);
		if (fp_out==NULL) throw ERR_OPEN;
		written = fwrite(&head_out,1,sizeof(head_out),fp_out);
		if (written<sizeof(head_out)) throw ERR_WRITE;
		written = fwrite(&palette,1,palsize_bytes,fp_out);
		if (written<palsize_bytes) throw ERR_WRITE;

		for (int y=0; y<head_in.ysize; y++) {
			read=fread(&buffer_in,1,rounded_scanline_in,fp_in);
			if (read<rounded_scanline_in) throw ERR_READ;
			if ( (y1<head_in.ysize-y) && (head_in.ysize-y<=y2) ) {
        bitcopy(buffer_out,buffer_in,0,x1*head_in.bpp,scanline_out_bits);
				written = fwrite(buffer_out,1,rounded_scanline_out,fp_out);
				if (written<rounded_scanline_out) throw ERR_WRITE;
			}
		}

		fclose(fp_in); fp_in=NULL;
		fclose(fp_out); fp_out=NULL;
	}
	catch (char *message) {
		printf("%s\n",message);
		if (fp_in!=NULL) fclose(fp_in);
		if (fp_out!=NULL) fclose(fp_out);
		return false;
	}
	return true;
}

int main(int argc, char *argv[],char *envp[]) {
	int x1,y1,x2,y2;
	if ((argc<6) ||
	    !sscanf(argv[3],"%d",&x1) ||
	    !sscanf(argv[4],"%d",&y1) ||
	    !sscanf(argv[5],"%d",&x2) ||
	    !sscanf(argv[6],"%d",&y2)) {
		printf("Usage: bmpcrop input.bmp output.bmp x1 y1 x2 y2\n");
		return 1;
	}

	if ((x2<=x1) || (y2<=y1)) {
		printf("(x1,y1) has to be lower than (x2,y2)\n");
		return 1;
	}

	if (!crop(argv[1],argv[2],x1,y1,x2,y2)) return 1;
   	return 0;
}
