
#include <stdio.h>
#include <string.h>

#define MAX_PALETTE     256 // entries
#define MAX_SCANLINE 131072 // bytes
#define MAX_TILES      1024 // horizontal

#define ERR_OPEN     "Error opening file"
#define ERR_READ     "Error reading file"
#define ERR_WRITE    "Error writing file"
#define ERR_NO_FILES "No files found"
#define ERR_NOT_READ "Bitmap info not read"

#pragma pack(push,1)
typedef struct {
	__int16 bm;
	__int32 fsize,reserved,dataofs,headsize,xsize,ysize;
	__int16 planes,bpp;
	__int32 compress,mapsize,xres,yres,palsize,palimp;
} BMPhead;
#pragma pack(pop)

#include <windows.h>
class Search {
private:
  HANDLE handle;
  const char* mask;
public:
  Search(const char* mask) {
    this->mask = mask;
    handle = INVALID_HANDLE_VALUE;
  }
  ~Search() {
    if (handle!=INVALID_HANDLE_VALUE) FindClose(handle);
    this->mask = mask;
  }
  const char* next() {
    WIN32_FIND_DATA result;
    if (handle==INVALID_HANDLE_VALUE) {
      handle = FindFirstFile(mask, &result);
      if (handle==INVALID_HANDLE_VALUE) return NULL;
    }
    else {
      if (!FindNextFile(handle, &result)) return NULL;
    }
    return result.cFileName;
  }
};

bool getnumbers(const char *str, int &x, int &y, bool minus) {
  char first[32], second[32];
  int digits = 0, chunks = 0, i = 0;
  int start[2], stop[2];
  bool last = false;
  while (str[i] && (chunks<2)) {
    bool digit = ('0'<=str[i]) && (str[i]<='9') || (minus && str[i]=='-');
    if (digit) digits++;
    if (digit && !last) start[chunks] = i;
    if (!digit && last) stop[chunks++] = i;
    last = digit;
    i++;
  }
  switch (chunks) {
    case 1:  i = (stop[0]-start[0])/2;
             memcpy(first, &(str[stop[0]-2*i]), i);
             first[i] = '\0';
             memcpy(second, &(str[stop[0]-i]), i);
             second[i] = '\0';
             break;
    case 2:  i = stop[0]-start[0];
             memcpy(first, &(str[start[0]]), i);
             first[i] = '\0';
             i = stop[1]-start[1];
             memcpy(second, &(str[start[1]]), i);
             second[i] = '\0';
             break;
    default: return false;
  }
  if (sscanf(first,"%d",&y)!=1) return false;
  if (sscanf(second,"%d",&x)!=1) return false;
  return true;
}

class Tile {
public:
  Tile(const char *name, int x, int y, Tile *next) {
    strcpy(this->name, name);
    this->x = x; this->y = y;
    this->next = next;
  }
  ~Tile() {
    if (next) delete next;
  }
  Tile *next;
  char name[256];
  int x, y;
};

void swap(int &a, int &b) {
  int t = a; a = b; b = t;
}

FILE* open(const char* name, const char* mode) {
  FILE* fp = fopen(name,mode);
  if (!fp) throw ERR_OPEN;
  return fp;
}

void close(FILE* fp) {
  if (fp) fclose(fp);
  fp = NULL;
}

void read(FILE* fp, void *buffer, int size) {
  if (fp) {
    int read = fread(buffer, 1, size, fp);
    if (read<size) throw ERR_READ;
  }
  else {
    memset(buffer, 0, size);
  }
}

void write(FILE* fp, void *buffer, int size) {
  int read = fwrite(buffer, 1, size, fp);
  if (read<size) throw ERR_WRITE;
}

int roundup_bits(int n) {
	return ((n+31)/32)*4;
}

void bitcopy(unsigned char *dst,const unsigned char *src,int dst_offset,int src_offset,int len) {
  int dst_int = dst_offset / 8 , src_int = src_offset / 8;
  int dst_mod = dst_offset % 8 , src_mod = src_offset % 8;
  int len_mid = (len + dst_mod) / 8 - 1;
  int shift = (src_mod - dst_mod) % 8;
  if (shift<0) { shift += 8; src_int--; }
  int inv_shift = 8 - shift;

  unsigned char mask1 = (0xFF>>dst_mod);
  unsigned char mask2 = (0xFF<<8-(dst_offset+len)%8);
  unsigned char byte1 = mask1 & ((src[src_int]<<shift) | (src[src_int+1]>>inv_shift));
  unsigned char byte2 = mask2 & ((src[src_int+len_mid+1]<<shift) | (src[src_int+len_mid+2]>>inv_shift));

  if (len_mid>=0) {
    dst[dst_int] = (~mask1 & dst[dst_int]) | byte1;
    for (int i=1; i<=len_mid; i++) {
      dst[dst_int+i] = (src[src_int+i]<<shift) | (src[src_int+i+1]>>inv_shift);
    }
    dst[dst_int+len_mid+1] = (~mask2 & dst[dst_int+len_mid+1]) | byte2;
  }
  else {
    dst[dst_int] = ~(mask1 & mask2) & dst[dst_int] | (byte1 & byte2);
  }
}

bool join(const char* mask, const char* out,int style,int overlap) {
  BMPhead head_in, head_out;
  unsigned char palette[4*MAX_PALETTE];
  int x,y,palsize_bytes;
  Tile** array = NULL;
  FILE *fp_out = NULL;
  Tile *tiles = NULL;
  FILE *fp_in[MAX_TILES]; memset(fp_in,0,MAX_TILES*sizeof(FILE*));
  static char errmsg[256];
  try {
    printf("Scanning .. ");
    Tile *act;
    int cnt = 0;
    Search search(mask);
    const char* name;
    while (name = search.next()) {
      int x, y;
      if (!getnumbers(name, x, y, true) && !getnumbers(name, x, y, false)) {
        sprintf(errmsg,"Invalid name (%s)",name);
        throw errmsg;
      }
      tiles = new Tile(name, x, y, tiles);
      cnt++;
    }
    printf("%d files found\n",cnt);
    if (cnt==0) throw ERR_NO_FILES;

    int miny, miny2, minx, minx2, maxx, maxy;
    miny = miny2 = minx = minx2 = 999999;
    maxx = maxy = -999999;
    for (act=tiles; act; act=act->next) {
      if (act->x < minx) {
        minx2 = minx;
        minx = act->x;
      }
      else if (act->x < minx2 && act->x > minx) {
        minx2 = act->x;
      }
      if (act->y < miny) {
        miny2 = miny;
        miny = act->y;
      }
      else if (act->y < miny2 && act->y > miny) {
        miny2 = act->y;
      }
      if (act->x > maxx) maxx = act->x;
      if (act->y > maxy) maxy = act->y;
    }

    int stepx = minx2 - minx;
    int stepy = miny2 - miny;

    printf("X coord = (%d..%d) step %d\n",minx,maxx,stepx);
    printf("Y coord = (%d..%d) step %d\n",miny,maxy,stepy);

    int sizex = (maxx - minx)/stepx + 1;
    int sizey = (maxy - miny)/stepy + 1;
    if (style & 0x04) swap(sizex,sizey);

    array = new Tile*[sizex*sizey];
    memset(array,0,sizex*sizey*sizeof(Tile*));
    int used = 0;
    for (act=tiles; act; act=act->next) {
      int x = (act->x-minx)/stepx;
      int y = (act->y-miny)/stepy;
      if ( (style & 0x04)) swap(x,y);
      if ( (style & 0x02)) x = sizex-x-1;
      if (!(style & 0x01)) y = sizey-y-1;
      if (!array[x+y*sizex]) used++;
      array[x+y*sizex] = act;
    }

    printf("Grid of %d x %d tiles, %d/%d used\n",sizex,sizey,used,sizex*sizey);

    bool inforead = false;
    for (y=0; y<sizey; y++) {
      for (x=0; x<sizex; x++) {
        if (array[x+y*sizex]) {
          printf("Reading bitmap info .. ");
          FILE* test = open(array[x+y*sizex]->name,"rb");
          read(test,&head_in,sizeof(head_in));
          palsize_bytes = head_in.dataofs-sizeof(head_in);
          read(test,palette,palsize_bytes);
          close(test);
          printf("%d x %d x %d bit - palette %d B\n",
            head_in.xsize,head_in.ysize,head_in.bpp,palsize_bytes);
          inforead = true;
          break;
        }
      }
      if (inforead) break;
    }
    if (!inforead) throw ERR_NOT_READ;

    head_out = head_in;
    head_out.xsize = head_in.xsize*sizex-overlap*(sizex-1);
    head_out.ysize = head_in.ysize*sizey-overlap*(sizey-1);
    int scanline_in_bits  = head_in.xsize*head_in.bpp;
    int scanline_out_bits = head_out.xsize*head_out.bpp;
    int scanline_out_step_bits = (head_in.xsize-overlap)*head_in.bpp;
    int overlapA = (overlap+1)/2;
    int overlapB = (overlap+0)/2;
    int overlapA_bits = overlapA*head_in.bpp;
    int rounded_scanline_in = roundup_bits(scanline_in_bits);
    int rounded_scanline_out = roundup_bits(scanline_out_bits);
    head_out.fsize = head_out.dataofs+head_out.ysize*rounded_scanline_out;

    printf("Composing ..    ");
    fp_out = open(out,"wb");
    write(fp_out,&head_out,sizeof(head_out));
    write(fp_out,palette,palsize_bytes);
    for (y=0; y<sizey; y++) {
      for (x=0; x<sizex; x++) {
        if (array[x+y*sizex]) {
          fp_in[x] = open(array[x+y*sizex]->name,"rb");
          fseek(fp_in[x],sizeof(head_in)+palsize_bytes,SEEK_SET);
        }
        else {
          fp_in[x] = NULL;
        }
      }
      for (int row=0; row<head_in.ysize; row++) {
        unsigned char all[MAX_SCANLINE];
        for (int x=0; x<sizex; x++) {
          unsigned char inbuf[MAX_SCANLINE];
          read(fp_in[x],inbuf,rounded_scanline_in);
          if (x==0)
            bitcopy(all,inbuf,0,0,scanline_in_bits);
          else
            bitcopy(all,inbuf,overlapA_bits+x*scanline_out_step_bits,overlapA_bits,scanline_in_bits);
        }
        if ((y==0 || row>=overlapA) && (y>=sizey-1 || row<head_in.ysize-overlapB))
          write(fp_out,all,rounded_scanline_out);
        int l = 100*(row+1+(y-1)*head_in.ysize)/(sizey*head_in.ysize);
        int p = 100*(row+1+(y+0)*head_in.ysize)/(sizey*head_in.ysize);
        if (p>l) printf("\b\b\b\b%3d%%",p);
      }
      for (int x=0; x<sizex; x++) close(fp_in[x]);
    }
    close(fp_out);
    printf("\n");
    delete array;
    delete tiles;
  }
  catch (char *message) {
    printf("%s\n",message);
    if (fp_out!=NULL) close(fp_out);
    for (int i=0; i<MAX_TILES; i++) if (fp_in[i]!=NULL) close(fp_in[i]);
    delete array;
    delete tiles;
    return false;
  }
  return true;
}

const char *usage = "Usage: bmpjoin [-style (0-7)] [-overlap N] input*.bmp output.bmp\n";

int main(int argc, char *argv[],char *envp[]) {
	int style = 0, overlap = 0;
	const char *input = NULL, *output = NULL;
	bool ok = false;
	for (int i=1; i<argc; i++) {
		if (strcmp(argv[i],"-style")==0) {
			if (++i>=argc || !sscanf(argv[i],"%d",&style)) break;
		}
		else if (strcmp(argv[i],"-overlap")==0) {
			if (++i>=argc || !sscanf(argv[i],"%d",&overlap)) break;
		}
		else {
			if (i+1>=argc) break;
			input  = argv[i];
			output = argv[i+1];
			ok = true;
			break;
		}
	}
	if (!ok) { printf(usage); return 1; }
	if (!join(input,output,style,overlap)) return 1;
	return 0;
}
